@@ -295,6 +295,12 @@
 ;; V1.79 07June2005 Jari Aalto <jari.aalto@cante.net>
 ;;  - fix byte-compilation warning about
 ;;    `(fboundp (quote imenu))' called for effect (Closes: #309788)
+;; V1.80 15Sep2005   Rafael Laboissiere <rafael@debian.org>
+;;  - Add debian-changelog-add-version-hook defaulting to
+;;    debian-changelog-add-new-upstream-release (Closes: #296725)
+;; V1.81 19Sep2005 Peter S Galbraith <psg@debian.org>
+;;  - Add outline-regexp and C-cC-n and C-cC-p movement commands as
+;;    suggested by Romain Francoise <rfrancoise@debian.org> (Closes: #322994)
 
 ;;; Acknowledgements:  (These people have contributed)
 ;;   Roland Rosenfeld <roland@debian.org>
@@ -361,6 +367,13 @@ If you do not wish this behaviour, reset it in your .emacs file like so:
   :type 'hook
   :options '(turn-on-auto-fill flyspell-mode))
 
+(defcustom debian-changelog-add-version-hook
+  (list 'debian-changelog-add-new-upstream-release)
+  "Hooks run just before inserting the signature separator \"--\" in a 
+new version in debian/changelog."
+  :group 'debian-changelog
+  :type 'hook)
+
 (defvar debian-changelog-local-variables-maybe-remove-done nil
   "Internal flag so we prompt only once.")
 
@@ -537,6 +550,9 @@ Upload to " val  " anyway?")))
 ;; keymap table definition
 ;;
 
+(autoload 'outline-next-visible-heading "outline")
+(autoload 'outline-prev-visible-heading "outline")
+
 (defvar debian-changelog-mode-map nil
   "Keymap for Debian changelog major mode.")
 (if debian-changelog-mode-map
@@ -557,7 +573,12 @@ Upload to " val  " anyway?")))
   (define-key debian-changelog-mode-map "\C-c\C-u"
     'debian-changelog-urgency)
   (define-key debian-changelog-mode-map "\C-c\C-e"
-    'debian-changelog-unfinalise-last-version))
+    'debian-changelog-unfinalise-last-version)
+  (define-key debian-changelog-mode-map "\C-c\C-n"
+    'outline-next-visible-heading)
+  (define-key debian-changelog-mode-map "\C-c\C-p"
+    'outline-previous-visible-heading))
+ 
 
 ;;
 ;; menu definition (Chris Waters)
@@ -831,6 +852,13 @@ for the debian/changelog file to add the entry to."
 ;;
 
 (defvar debian-changelog-new-upstream-release-p nil)
+
+(defun debian-changelog-add-new-upstream-release ()
+  "Normal hook for adding \"new upstream release\" entry to changelog."
+    (when debian-changelog-new-upstream-release-p
+      (insert "New upstream release")
+      (setq debian-changelog-new-upstream-release-p nil)))
+
 (defun debian-changelog-add-version ()
   "Add a new version section to a debian-style changelog file.
 If file is empty, create initial entry."
@@ -847,9 +875,7 @@ If file is empty, create initial entry."
     (if (debian-changelog-experimental-p)
         (insert pkg-name " (" version ") experimental; urgency=low\n\n  * ")
       (insert pkg-name " (" version ") unstable; urgency=low\n\n  * "))
-    (when debian-changelog-new-upstream-release-p
-      (insert "New upstream release")
-      (setq debian-changelog-new-upstream-release-p nil))
+    (run-hooks 'debian-changelog-add-version-hook)
     (save-excursion (insert "\n\n --\n\n"))))
 
 (defun debian-changelog-experimental-p ()
@@ -1209,6 +1235,7 @@ interface to set it, or simply set the variable
   (set (make-local-variable
 	'debian-changelog-local-variables-maybe-remove-done) nil)
   (set (make-local-variable 'indent-line-function) 'indent-relative-maybe)
+  (set (make-local-variable 'outline-regexp) "^[a-z]")
   (setq local-abbrev-table text-mode-abbrev-table)
   (set-syntax-table text-mode-syntax-table)
   (debian-bug-bug-menu-init debian-changelog-mode-map)
